function Distribution = Huber(TimeTrace,Kernel,L,RegularizationParameter,options)
%--------------------------------------------------------------------------
% Regularization of DEER data using the Huber penalty functional [1][2] as 
% well as the smoothed pseudo-Huber penalty functional [3][2]. These impose
% a L1-L2 penalty based on the value of the "Huber Parameter" which is
% optimally selected according to edge indicators. 
%
% The regularization problems have been implemented for the MATLAB fmincon 
% solver as well as for the convex optimization modeling framework CVX [4]
% (requires installation and license)
%
% Literature: 
% [1] Robust Estimation of a Location Parameter, P.J. Huber, Ann. Math. Statist. 
% [2] An Investigation of Smooth TV-Like Regularization in the Context of
%     the Optical Flow Problem, J. Imaging 2018, 4, 31
% [3] Multiple view geometry in computer vision, Cambridge University Press, MA, USA, 2003
% [4] Graph Implementations for Nonsmooth Convex Programs, M.C. Grant. P.S.
%     Boyd, Springer, London, 2008, 95-110
%
%--------------------------------------------------------------------------
% Luis Fabregas, 2018
%--------------------------------------------------------------------------

warning('on','all')
if ~isfield(options,'fminconTikhInitGuess')
 options.fminconTikhInitGuess = false;
end
if strcmp(options.Solver,'cvx')
  if exist('cvx','file')~=2
    warning('CVX not found. Make sure it is installed properly.')
    options.Solver = 'fmincon';
  end
end
warning('off','all')

%--------------------------------------------------------------------------
% Preparations
%--------------------------------------------------------------------------

%Adjustments
HuberParameter = options.HuberParameter;
SizeDistribution = length(TimeTrace);
TimeTrace = TimeTrace';
Distribution = zeros(SizeDistribution,1);

% Use SVD - Tikhonov solution as initial guess with fmincon solver (default deactivated, CVX does not use initial guesses) 
if options.fminconTikhInitGuess
[Distribution,~,~] = tikhonov(options.U,options.sm,options.X,TimeTrace',RegularizationParameter);
end

%Define and set options for fmincon solver (set to user-defined gradients)
SolverOptions = optimset('GradObj','on','MaxFunEvals',100000,'Display','off','MaxIter',100000);

%Non-negative constraint for fmincon
NonNegativeConstraint = zeros(SizeDistribution,1);

%--------------------------------------------------------------------------
% Call solvers
%--------------------------------------------------------------------------

switch options.Solver
  
  % fmincon - Pseudo-Huber
  %------------------------------------------------------------------------
  case 'fmincon'
    Distribution =  fmincon(@(Distribution)PseudoHuberFunctional(TimeTrace,Distribution.^2,L,Kernel,RegularizationParameter,options),Distribution,[],[],[],[],NonNegativeConstraint,[],[],SolverOptions);
    Distribution = Distribution.^2;
    
  % CVX - Pseudo-Huber
  %------------------------------------------------------------------------
  case 'cvx'
    cvx_begin quiet
        variable Distribution(SizeDistribution)
        Auxiliary = L*Distribution;
        % --------------------------------------------------------
        % Workaround for CVX: norm([x 1]) = sqrt(x^2 + 1) (convex)
        %     --> norm([x 1]) allowed by DCP rules
        %     --> sqrt(x^2 + 1) not allowed by DCP rules
        % --------------------------------------------------------
        clear VectorialNorm
        % Vectorial norm not implemented until MATLAB R2017b, do manually
        for i=1:length(Auxiliary)
          VectorialNorm(i) =  norm([Auxiliary(i)/HuberParameter 1]); % (Ignore preallocation warning to ensure 'cvx convex expression' class)
        end
        minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + RegularizationParameter^2*sum((VectorialNorm - 1)))
        subject to
            Distribution >= 0;
    cvx_end
    
end
    
end

%--------------------------------------------------------------------------
% fmincon - Objective function definitions
%--------------------------------------------------------------------------



% Pseudo-Huber Objective function (Functional + Gradient) for fmincon
%--------------------------------------------------------------------------
function [Functional,Gradient] = PseudoHuberFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,options)
%Get the pseudo Huber norm value
HuberParameter =  options.HuberParameter;
Auxiliary = L*Distribution;
HuberPenalty = (sqrt(1+(Auxiliary/HuberParameter).^2)-1);
HuberFunctional = sum(HuberPenalty);
%Get the Huber norm gradient
HuberGradient =  RegularizationParameter^2*(L/(HuberParameter))'*diag(1./sqrt((L*Distribution/HuberParameter).^2 + 1))*(L/(HuberParameter))*Distribution;
Functional = 1/2*norm(TimeTrace - Kernel*Distribution)^2 + RegularizationParameter^2*HuberFunctional;
Gradient = Kernel'*(Kernel*Distribution - TimeTrace) + RegularizationParameter^2*HuberGradient;
end
