function Distribution = OBIR(TimeTrace,Kernel,L,RegularizationParameter,options)
%--------------------------------------------------------------------------
% OSHER'S BREGMAN ITERATED REGULARIZATION (OBIR) METHOD 
%--------------------------------------------------------------------------
% Regularization of DEER data by Bregman iterations [1] equiped with different
% penalty functionals of choice [2]. Bregman iterations allow the recovery of
% noiseless signal from the residuals of the previous iteration. The
% optimal solution is determined to be found when the residual standard
% deviation is equal or less than the noise standard deviation [3]. Therefore
% the level of noise given as an input is determinant for the succesfull
% performance of the method. Nontheless, when the noise level can be well
% approximated, this method will yield better results than any other
% non-Bregman-iterated method even with optimal regularization parameter
% choice [4]. 
%
% Literature:
% [1] L.M. Bregman, USSR Computational Mathematics and Mathematical Physics (1967) 7, 3, 200-217
% [2] Charest et al, 2006 40th Annual Conference on Information Sciences and Systems
% [3] Fabregas and Jeschke, Journal Magnetic Resonance (2018)
% [4] Yin et al, SIAM Journal on Imaging Sciences (2008) 1, 143-168
%
%--------------------------------------------------------------------------
% L. Fabregas, 2018
%--------------------------------------------------------------------------

warning('on','all')
if ~isfield(options,'plotBregmanIterations')
  options.plotBregmanIterations = false;
end
if ~isfield(options,'StopIfBregmanDiverge')
  options.StopIfBregmanDiverge = false;
end
if ~isfield(options,'Solver')
  options.Solver = 'fmincon';
end
if ~isfield(options,'HuberParameter')
  warning('Huber parameter was not defined for OBIR in options. Setting to default: 1.35')
  options.HuberParameter = 1.35;
end
if ~isfield(options,'PenaltyFunctional')
  warning('Penalty was not defined for OBIR in options. Setting to default: TIKH')
  options.PenaltyFunctional = 'Tikh';
end
if ~isfield(options,'NoiseLevel')
  error('Error: Noise level was not defined for OBIR in options.')
end
if strcmp(options.Solver,'cvx')
  if exist('cvx','file')~=2
    warning('CVX not found. Make sure it is installed properly.')
    options.Solver = 'fmincon';
  end
end
warning('off','all')

%--------------------------------------------------------------------------
% Preparation
%--------------------------------------------------------------------------

%Initialize 
SizeDistribution = length(TimeTrace);
Subgradient = zeros(SizeDistribution,1);
Counter = 1;
Iteration = 1;
TimeTrace = TimeTrace';

Distribution = zeros(SizeDistribution,1);
%For some reason, zeros as initial point for fmincon crashes
if strcmp(options.Solver,'fmincon')  
  [Distribution,~,~] = tikhonov(options.U,options.sm,options.X,TimeTrace,RegularizationParameter);
  %Set solver options for fmincon solver
  SolverOptions = optimset('MaxFunEvals',100000,'Display','off','MaxIter',100000);
  %Non-negative constraint for fmincon solver
  NonNegativeConstraint = zeros(1,length(TimeTrace));
end

%--------------------------------------------------------------------------
% Osher's Bregman Iterations Algorithm
%--------------------------------------------------------------------------

while Iteration <= 200
  CheckDistribution = Distribution;
  %Select the desired Osher functional and solve
  switch options.PenaltyFunctional
    
    % Total Variation Penalty
    %--------------------------------------------------------------------------
    case 'TV'
      switch options.Solver
        case 'fmincon'
          [Distribution,CostValue(Iteration)] =  fmincon(@(Distribution)TVOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,Subgradient),Distribution,[],[],[],[],NonNegativeConstraint,[],[],SolverOptions);
        case 'cvx'
          SizeDistribution = length(TimeTrace);
          cvx_begin quiet
              variable Distribution(SizeDistribution)
              clear VectorialNorm
              Auxiliary = L*Distribution;
              % Vectorial norm has not implemented until MATLAB R2017b, do manually
              for i=1:length(Auxiliary)
                %Ignore preallocation warning to ensure 'cvx convex expression' class
                VectorialNorm(i) =  norm([Auxiliary(i) 1e-12]);
              end
              minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + RegularizationParameter^2*sum(VectorialNorm)  + dot(Distribution,Subgradient) )
              subject to
                  Distribution >= 0;
          cvx_end
      end
      
    % Tikhonov Penalty
    %--------------------------------------------------------------------------
    case 'Tikh'
      switch options.Solver
        case 'fmincon'
          [Distribution,CostValue(Iteration)] =  fmincon(@(Distribution)TikhonovOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,Subgradient),Distribution,[],[],[],[],NonNegativeConstraint,[],[],SolverOptions);
        case 'cvx'
          cvx_begin quiet
              variable Distribution(SizeDistribution)
              minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + square_pos(norm(RegularizationParameter*L*Distribution,'fro')) + dot(Distribution,Subgradient))
              subject to
                 Distribution >= 0;
          cvx_end
          CostValue(Iteration) = cvx_optval;
      end
      
    % Huber Penalty 
    %--------------------------------------------------------------------------  
    case 'Huber'
      switch options.Solver
        case 'fmincon'
          [Distribution,CostValue(Iteration)] =  fmincon(@(Distribution)PseudoHuberOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,Subgradient,options),Distribution,[],[],[],[],[],[],[],SolverOptions);
          Distribution = Distribution.^2;
        case 'cvx'
          SizeDistribution = length(TimeTrace);
          HuberParameter = options.HuberParameter;
          clear VectorialNorm
          cvx_begin quiet
                variable Distribution(SizeDistribution)
                Auxiliary = L*Distribution;
                % --------------------------------------------------------
                % Workaround for CVX: norm([x 1]) = sqrt(x^2 + 1) (convex)
                %     --> norm([x 1])   allowed by DCP rules
                %     --> sqrt(x^2 + 1) not allowed by DCP rules
                % --------------------------------------------------------
                clear VectorialNorm
                % Vectorial norm not implemented until MATLAB R2017b, do manually
                for i=1:length(Auxiliary)
                    VectorialNorm(i) =  norm([Auxiliary(i)/HuberParameter 1]);
                end
                minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + RegularizationParameter^2*sum((VectorialNorm - 1)) +  dot(Distribution,Subgradient))
                subject to
                    Distribution >= 0;
          cvx_end
          CostValue(Iteration) = cvx_optval; 
      end
      
    % Circular Symmetric Huber Penalty
    %--------------------------------------------------------------------------
    case 'Hubercirc'
      SizeDistribution = length(TimeTrace);
      HuberParameter =  options.HuberParameter;
      cvx_begin quiet
      variable Distribution(SizeDistribution)
      minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + 100*RegularizationParameter^2*(huber_circ(L*Distribution,[],HuberParameter)) +  dot(Distribution,Subgradient))
      subject to
      Distribution >= 0;
      cvx_end
      CostValue(Iteration) = cvx_optval;
      
    % Inverse Huber Penalty
    %--------------------------------------------------------------------------        
    case 'Berhu'
      switch options.Solver
        case 'fmincon'
          [Distribution,CostValue(Iteration)] =  fmincon(@(Distribution)BerhuOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,Subgradient,options),Distribution,[],[],[],[],NonNegativeConstraint,[],[],SolverOptions);
        case 'cvx'
            SizeDistribution = length(TimeTrace);
            HuberParameter = options.HuberParameter;
            cvx_begin quiet
            variable Distribution(SizeDistribution)
            minimize( 1/2*square_pos(norm(Kernel*Distribution - TimeTrace,'fro')) + RegularizationParameter^2*sum(berhu(L*Distribution,HuberParameter)) + dot(Distribution,Subgradient))
            subject to
            Distribution >= 0;
            cvx_end
            CostValue(Iteration) = cvx_optval;
      end
  end
    
  %Update subgradient at current solution
  Subgradient = Subgradient + Kernel'*(Kernel*Distribution - TimeTrace);
  
  NoiseDeviation = options.NoiseLevel;
  
   Distribution = Distribution/sum(Distribution);
    if options.plotBregmanIterations
      ResidualDeviation(Iteration) = std(Kernel*Distribution - TimeTrace);
      figure(4124),set(gcf,'NumberTitle','off','Name','Bregman Iterations','Units','pixels','Position',[186 513 1055 421]),clf
      subplot(1,2,1),plot(Distribution/sum(Distribution),'k','LineWidth',1.5)
      set(gca,'Ytick',[],'Xtick',[]),title('Distribution')
      subplot(122),plot(1:Iteration,ResidualDeviation,'ko--','LineWidth',1.5);hold on,plot(1:Iteration+1,ones(Iteration+1,1)*options.NoiseLevel,'r--','LineWidth',1.5);
      xlabel('Bregman Iteration'),ylabel('Residual Std. Dev.'),ylims = ylim; ylim([0.95*ylims(1) ylims(2)]),title('Functional'),drawnow,
    end
  
  %--------------------------------------------------------------------------
  %Iteration Control
  %-------------------------------------------------------------------------- 
  if Iteration == 1
    %If at first iteration, the residual deviation is already below the noise deviation then impose oversmoothing and remain at first iteration
    if NoiseDeviation  > std(Kernel*Distribution - TimeTrace)
      RegularizationParameter = RegularizationParameter*2^Counter;
      Counter = Counter + 1;
    else
      %Once the residual deviation is above the treshold, then proceed further with the Bregman iterations
      Iteration  = Iteration +1;
    end
  else
    %For the rest of the Bregman iterations then control the condition and stop when fulfilled
    if NoiseDeviation  > std(Kernel*Distribution - TimeTrace)
      break;
    else
      Iteration  = Iteration +1;
    end
    %If residual deviation starts to diverge, stop
    if std(Kernel*CheckDistribution - TimeTrace) < std(Kernel*Distribution - TimeTrace) && options.StopIfBregmanDiverge
      Distribution = CheckDistribution;
      break;
    end
  end
  
end

end

%--------------------------------------------------------------------------
% Funtional Definitions for fmincon Solver
%--------------------------------------------------------------------------

% Osher-Tikhonov
function [Functional] = TikhonovOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,PreviousSubgradient)
Functional = 1/2*norm(TimeTrace - Kernel*Distribution)^2 + norm(RegularizationParameter*(L*(Distribution)))^2 + dot(Distribution,PreviousSubgradient);
end

% Osher-TV
function [Functional] = TVOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,PreviousSubgradient)
Functional = 1/2*norm(TimeTrace - Kernel*Distribution)^2 + RegularizationParameter^2*sum(sqrt(abs(L*(Distribution)).^2 + 1e-12)) + dot(Distribution,PreviousSubgradient);
end

 %Osher-Huber
function [Functional] = PseudoHuberOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,Subgradient,options)
%Get the pseudo Huber norm value
HuberParameter = options.HuberParameter;
Auxiliary = L*Distribution.^2;
HuberLoss = sqrt(1+(Auxiliary/HuberParameter).^2) - 1;
HuberFunctional = sum(HuberLoss);
Functional = 1/2*norm(TimeTrace - Kernel*Distribution.^2)^2 + RegularizationParameter^2*HuberFunctional + dot(Distribution.^2,Subgradient);

end


function [Functional] = BerhuOsherFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter,PreviousSubgradient,options)
%Get the Berhu norm value
HuberParameter = options.HuberParameter;
Auxiliary = L*Distribution;
BerhuLoss = zeros(length(Auxiliary),1);
for i=1:length(Auxiliary)
  if abs(Auxiliary(i)) > HuberParameter
    BerhuLoss = (Auxiliary(i)^2 - HuberParameter^2)/(2*HuberParameter);
  else
    BerhuLoss = abs(sqrt((Auxiliary(i))^2 + 1e-24));
  end
end
BerhuFunctional = sum(BerhuLoss);
Functional = 1/2*norm(TimeTrace - Kernel*Distribution)^2 + (sqrt(2) - 1)*RegularizationParameter^2*BerhuFunctional +  dot(Distribution,PreviousSubgradient);

end