function Distribution = Tikhonov2(TimeTrace,Kernel,L,RegularizationParameter,options)
%--------------------------------------------------------------------------
% Tikhonov regularization solved using the fmincon or CVX solvers instead
% of the standard lsnonneg used in DeerAnalysis. This function serves as a
% comparison to the DeerAnalysis function. All Tikhonov regularization in
% the paper [1] are computed using this function.
%
% [1] Luis Fabregas Ibanez, Gunnar Jeschke,
%     General regularization framework for DEER spectroscopy 
%     Journal of Magnetic Resonance, 2018
%--------------------------------------------------------------------------
% L. Fabregas, 2018
%--------------------------------------------------------------------------
warning('on','all')
if strcmp(options.Solver,'cvx')
  if exist('cvx','file')~=2
    warning('CVX not found. Make sure it is installed properly.')
    options.Solver = 'fmincon';
  end
end
warning('off','all')

%--------------------------------------------------------------------------
% Preparation
%--------------------------------------------------------------------------

%Non-negative constraint
NonNegativeConstraint = zeros(1,length(TimeTrace));
TimeTrace = TimeTrace';
if strcmp(options.Solver,'fmincon')
      Distribution = ones(length(TimeTrace),1);
end

%--------------------------------------------------------------------------
% Regularization
%--------------------------------------------------------------------------
switch options.Solver
   case 'cvx'
     SizeDistribution = length(TimeTrace);
     cvx_begin quiet
        variable Distribution(SizeDistribution)
        minimize( 1/2*square_pos(norm(Kernel*Distribution-TimeTrace,'fro')) + 1/2*square_pos(norm(RegularizationParameter*L*Distribution,'fro')))
        subject to 
            Distribution >= 0;
     cvx_end
  case 'fmincon'
      SolverOptions = optimoptions(@fmincon,'GradObj','off','Display','off','MaxFunEvals',100000,'MaxIter',100000);
      Distribution =  fmincon(@(Distribution) TikhonovFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter),Distribution,[],[],[],[],NonNegativeConstraint,[],[],SolverOptions);
end

end

function [Functional] = TikhonovFunctional(TimeTrace,Distribution,L,Kernel,RegularizationParameter)
Functional = 1/2*norm(Kernel*Distribution - TimeTrace)^2 + 1/2*norm(RegularizationParameter*L*Distribution)^2;
end
