%--------------------------------------------------------------------------
% DEMO FOR REGULARIZATION TOOLBOX 
% Luis Fabregas, 2018
%--------------------------------------------------------------------------
% This script serves as a template for the input of the regularizeDeer
% function. All the different options available are collected here along an
% example DEER time trace from the Edwards-Stoll library. This time trace
% corresponds to one of the distance distributions presented in the paper
% [1] as examples. 
% For more information about the toolbox the user can open the readme file
% containing the description in regularizeDeer.m or just type 
%    help regularizeDeer 
% in the Matlab console.
%
% [1] Luis Fabregas Ibanez, Gunnar Jeschke,
%     General regularization framework for DEER spectroscopy 
%     Journal of Magnetic Resonance, 2018
%--------------------------------------------------------------------------
clearvars

%Define number of figure to plot output
FigureNumber = 2;

%Load example from paper
load('Demo_TimeTrace.mat')

% Regularization method
%--------------------------------------------------------------------------
RegularizationMethod = 'obir';
options.Derivative = 2;
options.Solver = 'cvx';

% OBIR options
%--------------------------------------------------------------------------
options.PenaltyFunctional = 'Tikh';
options.StopIfBregmanDiverge = false;

% Regularization parameter selection options
%--------------------------------------------------------------------------
options.SelectionMethod = 'gml';
options.unconstrainedProblem = 1;

ForceRegularizationParameter = false;
if ForceRegularizationParameter
options.ForcedRegularizationParameter = 100;
options.ForcedHuberParameter = 1.262e-4;
end

% Display options
%--------------------------------------------------------------------------
%Plot distribution at each Bregman iteration
options.plotBregmanIterations = true;
%Plot the selection functional and optimal regularization parameter
options.plotSelectionFunctional = true;
%Plot which parts of distribution are L1/L2-weighted (for Huber)
options.plotL1L2regions = true;

% Set other options
%--------------------------------------------------------------------------
options.NoiseLevel = 0.0529;

% Launch regularization
%--------------------------------------------------------------------------
[Distribution,DistanceAxis,FittedTimeTrace] = regularizeDeer(TimeTrace,TimeAxis,RegularizationMethod,options);

%Normalize distribution to unity integral
Distribution = Distribution/sum(Distribution) ;


% Process output and plot
%--------------------------------------------------------------------------
figure(FigureNumber),set(gcf,'NumberTitle','off','Name','Regularization results','Units','pixels','Position',[186 513 1055 421]),clf
subplot(1,2,1)
plot(DistanceAxis,ModelDistributions,'LineWidth',1.5,'Color','k'),axis tight
xlabel('Distance [nm]'),ylabel('PDF'),drawnow
if options.unconstrainedProblem
Unconstrained = 'u';
else
Unconstrained = '';
end
title(sprintf('Distribution with %s (%s%s - L_{%i})',upper(RegularizationMethod),upper(options.SelectionMethod),Unconstrained,options.Derivative))
hold on
plot(DistanceAxis,Distribution,'LineWidth',1.5,'Color','r')
legend('True','Fit')
subplot(1,2,2)
plot(TimeAxis,TimeTrace,'LineWidth',1.5,'Color','k');
hold on;
plot(TimeAxis,FittedTimeTrace,'LineWidth',1.5,'Color','r'),title('Input form factor and fit'),legend('Input','Fit')
xlabel('Time [\mus]'),ylabel('Form Factor')
