function Indicator = edgeIndicator(x,L,Method)
%--------------------------------------------------------------------------
% An edge is a boundary between two uniform regions. You can detect an edge 
% by comparing the intensity of neighboring points. However, because uniform
% regions are not crisply defined, small intensity differences between two 
% neighboring points do not always represent an edge. [1]
% This function runs a simple characterization of edges in 1D-data sets by
% either considering the gradient/curvature or by using the
% fuzzy-interference ruled by else-action (FIRE) operator as a fuzzy edge
% indicator.[2]
% The fuzzy logic approach for allows you to use membership functions to 
% define the degree to which a point belongs to an edge or a uniform region.
%
% Literature:
% [1] Fuzzy Logic Image Processing, Fuzzy Logic Toolbox Documentation,
%     MathWorks
% [2] Image Restoration with Fuzzy Coefficient Driven Anisotropic Diffusion
%     B.K. Panigrahi et al. SEMCCO 2014, LNCS 8947, pp. 145155, 2015
%
%--------------------------------------------------------------------------
% L. Fabregas, 2018
%--------------------------------------------------------------------------
% L1 = gradient    L2 = curvature

Lx = L*x;

switch Method
  case 'gradient'
    Indicator = abs(Lx);
  case 'fire'
    NeighborhoodSize = 5;
    FireOperator = zeros(size(Lx));
    for i=NeighborhoodSize+1:length(Lx)-NeighborhoodSize-1
      Neighborhood = -NeighborhoodSize:NeighborhoodSize;
        FireOperator(i) = 1 - 1/(1 + sum(abs(Lx(i) - Lx(i + Neighborhood)))/numel(Neighborhood));
    end
    FireOperator = FireOperator/max(FireOperator);
    Indicator = abs(Lx.*FireOperator);
end