%------------------------------------------------------------------------------------------------
% GENERAL FRAMEWORK FOR THE REGULARIZATION OF DEER SPECTROSCOPY
%------------------------------------------------------------------------------------------------
% INPUT
%------------------------------------------------------------------------------------------------
%
% TimeTrace                         Form factor obtained from DEER/PELDOR experiment
%
% TimeAxis                          Time axis of the form factor in microseconds [us]
%
% RegularizationMethod              String corresponding to method to be employed
%       'tikhonov'                  Tikhonov regularization (same as DeerAnalysis2016)
%       'tikhonov2'                 Tikhonov regularization 
%       'tv'                        Total variation regularization 
%       'huber'                     Pseudo-Huber regularization 
%       'obir'                      Osher's Bregman iterated regularization
%
% options                           Structure containing all other additional parameters
%
%  .PenaltyFunctional               Determines the penalty for the OBIR method. Can be: 
%                                   'Tikh','Huber' ot 'TV'   
%
%  .Derivative                      Order of derivative to be used as regularization
%                                   matrix
%
%  .Solver                          Sets the solver to be used:
%                                   'fmincon' Internal Matlab solver (default)
%                                   'cvx' CVX solver (requires installation, which is free)
%
%  .SelectionMethod                 Functional for the selection of the regularization
%                                   parameter. Can be: 'gml','mcl','ncp','aic','aicc',
%                                   'bic','gcv','rgcv','srgcv','lr','lc','ee','rm','cv','dp'
%
%  .unconstrainedProblem            Constrained or unconstrained selection (true or
%                                   false)
%
%  .StopIfBregmanDiverge            In OBIR, stop the iterations if residual increases
%                                   instead of decreasing (true or false)
%
%  .NoiseLevel                      Approximation of the noise deviation in the time
%                                   trace to be used for OBIR
%
%  .ForcedRegularizationParameter   If given, the regularization parameter
%   &                               selection will be skipped and employ 
%  .ForcedHuberParameter            these values
%
%------------------------------------------------------------------------------------------------
% OUTPUT
%------------------------------------------------------------------------------------------------
%
% Distribution                      Regularized distance distribution (not normalized)
%
% DistanceAxis                      Corresponding distance axis in nanometers [nm]
%
% FittedTimeTrace                   Form factor corresponding to the output distance distribution
%
%------------------------------------------------------------------------------------------------
% NOTES
%------------------------------------------------------------------------------------------------
%
% - In case you want to use the CVX solver (highly recommended for speed) 
%   it has to be downloaded and installed from:
%                     http://cvxr.com/cvx/download/
%   The standard bundle of CVX is enough to run this package and no
%   commercial solver is required (although can be for some speed-up). Once
%   installed we recommend to add the following lines to startup.m
%     %Add CVX path
%     addpath(genpath('C:/Users/luis/Documents/MATLAB/cvx'))
%     %Add missing CVX path (a known bug)
%     addpath(genpath('C:/Users/luis/Documents/MATLAB/cvx/functions/vec_'))
%
% - This package was developed using MATLAB R2016b v.9.1. We cannot
%   guarantee the compatibility all functions with older versions.
%
% If you employ this or any of the contained functions please cite:
%
% [1] Luis Fabregas Ibanez, Gunnar Jeschke,
%     General regularization framework for DEER spectroscopy 
%     Journal of Magnetic Resonance, 2018
%
%--------------------------------------------------------------------------
% Luis Fabregas, 2018
%--------------------------------------------------------------------------

function [Distribution,DistanceAxis,FittedTimeTrace] = regularizeDeer(TimeTrace,TimeAxis,RegularizationMethod,options)

if ~isfield(options,'Derivative')
  warning('Derivative order not given as field in options. Using default: 2')
  options.Derivative = 2;
end
if ~isfield(options,'plotSelectionFunctional')
  options.plotSelectionFunctional = false;
end
if ~isfield(options,'SelectionMethod')
  warning('Selection method not specified as field in options. Using default: AIC')
  options.SelectionMethod = 'aic';
end
if ~isfield(options,'unconstrainedProblem')
  warning('Selection method constraint not specified as field in options. Using default: unconstrained')
  options.unconstrainedProblem = 1;
end
if ~isfield(options,'Solver')
  options.Solver = 'fmincon';
end

%--------------------------------------------------------------------------
% Preparations
%--------------------------------------------------------------------------
TimeStep =  TimeAxis(2) - TimeAxis(1);
Dimension= length(TimeTrace);
TimeTrace = real(TimeTrace);
if  iscolumn(TimeTrace)
TimeTrace = TimeTrace';
end
% Kernel Construction & SVD
[Kernel,DistanceAxis,~,U,sm,X,V,L] = get_kernel(Dimension,TimeStep,[],[],options.Derivative);
options.sm = sm;
options.X = X;
options.V = V;
options.U = U;

%--------------------------------------------------------------------------
% Regularization Parameter
%--------------------------------------------------------------------------

%Check which penalty is to be used
switch RegularizationMethod
  case {'tikhonov','tikhonov2'}
    options.SelectionPenaltyFunctional = 'Tikh';
  case 'tv'
    options.SelectionPenaltyFunctional = 'TV';
  case 'huber'
    options.SelectionPenaltyFunctional = 'Huber';
  case 'obir'
    switch options.PenaltyFunctional
      case 'TV'
        options.SelectionPenaltyFunctional = 'TV';
      case 'Tikh'
        options.SelectionPenaltyFunctional = 'Tikh';
      case 'Huber'
        options.SelectionPenaltyFunctional = 'Huber';
    end
  otherwise
    error('Error: No matching regularization method was found. Check your input for RegularizationMethod or options.PenaltyFunctional')
end

% If using Huber, then do quickly Tikhonov to get Huber Parameter
if strcmp(options.SelectionPenaltyFunctional,'Huber')
  options.SelectionPenaltyFunctional = 'Tikh';
  if options.plotSelectionFunctional
    Logical = options.plotSelectionFunctional;
  options.plotSelectionFunctional = false;
  end
  %Lambda-selection for Tikhonov
  [RegularizationParameter,~] = selectionMethods(options.SelectionMethod,TimeTrace,Kernel,L,U,sm,options);
  if Logical
    options.plotSelectionFunctional = Logical;
  end
  %Restore option
  options.SelectionPenaltyFunctional = 'Huber';
  %Fast Tikhonov regularization
  [Distribution,~,~] = tikhonov(U,sm,X,TimeTrace',RegularizationParameter);
  Distribution(Distribution<0)=0;
  C = [Kernel;RegularizationParameter*L];
  d = [TimeTrace';L*Distribution];
  SolverOptions = optimset('MaxFunEvals',100000,'Display','off','MaxIter',100000);
  Distribution = lsqnonneg(C,d,SolverOptions);
  %Compute the first estimate of the Huber parameter
  options.HuberParameter = mean(edgeIndicator(Distribution,L,'gradient'));
end

%Check if regularization or Huber parameters are to be enforced and skip selection if necessary
if isfield(options,'ForcedRegularizationParameter') && isfield(options,'ForcedHuberParameter')
  RegularizationParameter = options.ForcedRegularizationParameter;
  HuberParameter = options.ForcedHuberParameter;
elseif  isfield(options,'ForcedRegularizationParameter')
  [~,HuberParameter] = selectionMethods(options.SelectionMethod,TimeTrace,Kernel,L,U,sm,options);
  RegularizationParameter = options.ForcedRegularizationParameter;
else
  [RegularizationParameter,HuberParameter] = selectionMethods(options.SelectionMethod,TimeTrace,Kernel,L,U,sm,options);
end

%--------------------------------------------------------------------------
% Regularization Methods
%--------------------------------------------------------------------------
switch RegularizationMethod

% Tikhonov
%--------------------------------------------------------------------------    
  case 'tikhonov2'
    Distribution = Tikhonov2(TimeTrace,Kernel,L,RegularizationParameter,options);
    
% Tikhonov (DeerAnalysis)
%--------------------------------------------------------------------------    
  case 'tikhonov'
    [Distribution,~,~] = tikhonov(U,sm,X,TimeTrace',RegularizationParameter);
    Distribution(Distribution<0)=0;
    SolverOptions = optimset('MaxFunEvals',100000,'Display','off','MaxIter',100000);
    C = [Kernel;RegularizationParameter*L];
    d = [TimeTrace';L*Distribution];
    Distribution = lsqnonneg(C,d,SolverOptions); 

  case 'tv'
% Total Variation (TV)
%--------------------------------------------------------------------------     
    Distribution = TV(TimeTrace,Kernel,L,RegularizationParameter,options);  

% Osher's Bregman Iterative Regularization (OBIR)
%--------------------------------------------------------------------------       
  case 'obir'
    options.HuberParameter = HuberParameter;
    Distribution = OBIR(TimeTrace,Kernel,L,RegularizationParameter,options);

% pseudo-Huber Regularization
%--------------------------------------------------------------------------
  case 'huber'
    options.HuberParameter = HuberParameter;
    Distribution = Huber(TimeTrace,Kernel,L,RegularizationParameter,options);
  
%--------------------------------------------------------------------------    
  otherwise 
    error('Error: No matching regularization method was found. Check your input.')
end

% Time Trace Conversion
%--------------------------------------------------------------------------
try
  FittedTimeTrace = get_ff(DistanceAxis,Distribution',TimeAxis,Kernel');
catch 
  FittedTimeTrace = [];
end

end
