function [OptimalRegularizationParameter,OptimalHuberParameter] = selectionMethods(SelectionMethod,TimeTrace,Kernel,L,U,sm,options)
%--------------------------------------------------------------------------
% Find optimal regularization parameter for inversion of DEER data in order
% to obtain distance distributions. The selection of the optimal parameter
% value is performed according to a selection method of choice [1]. The
% function allows constrained and unconstrained parameter selection for
% different penalty functionals (Tikhonov [1], TV [2] and pseudo-Huber)
% The recursive nature of the gradients and pseudoinverses of non-Tikhonov
% penalties is handled as described in [3].
%
% Literature:
% [1] Optimal Tikhonov regularization for DEER spectroscopy, Edwards &
%     Stoll, JMR 288, 2018, 58-68
% [2] UPRE method for total variation parameter selection, Y.Lin, B.Wohlberg
%     H. Guo, Signal Processing 90, 2010, 2546-2551
% [3] General regularization framework for DEER spectroscopy 
%     L. Fabregas Ibanez,G. Jeschke, Journal of Magnetic Resonance, 2018 
%
%--------------------------------------------------------------------------
% Luis Fabregas, 2018
%--------------------------------------------------------------------------
warning('off','all')
%Set options defaults
if ~isfield(options,'SelectionPenaltyFunctional')
 options.SelectionPenaltyFunctional = 'Tikh';
end
if ~isfield(options,'LambdaParameterRange')
 options.LambdaParameterRange = 2;
end
if ~isfield(options,'plotLcurve')
 options.plotLcurve = 0;
end 
if ~isfield(options,'plotL1L2regions')
 options.plotL1L2regions = 0;
end 
if ~isfield(options,'HuberParameter')
 options.HuberParameter = [];
 HuberParameter = 1.35;
else
  HuberParameter = options.HuberParameter;
end
if ~isfield(options,'nPoints')
 options.nPoints = 150; % Points for regularization parameters array
end 


%--------------------------------------------------------------------------
% Preparations
%--------------------------------------------------------------------------

%Adjustments
L = full(L);
TimeTrace = TimeTrace';
DipolarDimension = length(TimeTrace);

if options.LambdaParameterRange == 1
% Construction of SVD-terms
[Dimension1,Dimension2] = size(sm);
BetaFactor = U'*TimeTrace;
if (Dimension2 == 1)
  SingularValues = sm;
  BetaFactor = BetaFactor(1:Dimension1);
else
  SingularValues = sm(Dimension1:-1:1,1)./sm(Dimension1:-1:1,2);
  BetaFactor = BetaFactor(Dimension1:-1:1);
end
xi = BetaFactor(1:Dimension1)./SingularValues;
end

%Construct appropiate regularization parameter vectors
switch options.LambdaParameterRange
  case 1
    %range type 1 (DeerAnalysis2016)
    nPoints = 100;
    ParameterMinimalRatio = 16*eps*1e6;
    RegularizationParameter(nPoints) = max([SingularValues(Dimension1),SingularValues(1)*ParameterMinimalRatio]);
    ParameterRatio = (SingularValues(1)/RegularizationParameter(nPoints))^(1/(nPoints-1));
    for i=nPoints-1:-1:1
      RegularizationParameter(i) = ParameterRatio*RegularizationParameter(i+1);
    end
  case 2
    %range type 2 (1dB scaling)
    RegularizationParameter = get_regparamrange(Kernel,L);
end

%Update number of points just to make sure 
nPoints = length(RegularizationParameter);
%Initialize arrays
Residual = zeros(1,nPoints);
Penalty = zeros(1,nPoints);
Functional  =zeros(1,nPoints);
%Initialize cells
PseudoInverse = cell(1,nPoints);
Distribution = cell(1,nPoints);
InfluenceMatrix = cell(1,nPoints);

%--------------------------------------------------------------------------
% Pseudo-Inverses and Distributions
%--------------------------------------------------------------------------

for i=1:nPoints %Loop over all regularization parameter values

  switch options.SelectionPenaltyFunctional
    
    %--------------------------------------------------------------------------
    % Total variation (L1) Penalty
    %--------------------------------------------------------------------------
    case 'TV'
 
      %Unconstrained distributions required for construction of correct pseudoinverse
      Distribution{i} = zeros(DipolarDimension,1);
      try
        %Define options for fsolve
        Solveroptions = optimoptions(@fsolve,'Display','off','Algorithm','trust-region-reflective');
        Distribution{i} = fsolve(@(Distribution) Kernel'*(Kernel*Distribution - TimeTrace) + ...
          RegularizationParameter(i)^2*(L)'*((L*Distribution)./sqrt((L*Distribution).^2 + 1e-24)) ...
          ,Distribution{i},Solveroptions);
        
        % Compute pseudoinverse by analytical expression (only defined for unconstrained distribution)
        localTVPseudoInverse = L'*((L./sqrt((L*Distribution{i}).^2 + 1e-24)));
        PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*localTVPseudoInverse)\Kernel';
        
        % If constrained solution is requested
        if ~options.unconstrainedProblem
          %Now that we have the pseudoinverse, the constrained distribution can be computed quickly
          Distribution{i} = fnnls(Kernel'*Kernel + RegularizationParameter(i)^2*localTVPseudoInverse,Kernel'*TimeTrace);
        end
      catch
        % Worst-case scenario when fsolve crashes, use this bad method to get at least something
        PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*(L')*L)\Kernel';
        Distribution{i}  = PseudoInverse{i}*TimeTrace; %unconstrained Tikhonov solution
        try
          localTVPseudoInverse = L'*diag(L./sqrt((L*Distribution{i}).^2 + 1e-24));
        catch
          localTVPseudoInverse = zeros(DipolarDimension,1);
        end
        PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*localTVPseudoInverse)\Kernel';
        Distribution{i}  = PseudoInverse{i}*TimeTrace;
      end

      
      %Get last variables required for the selection functionals
      Penalty(i) = norm(L*Distribution{i});
%       Residual defined this way so that later Residual(i)^2 equals 1/2*norm(...)^2
      Residual(i) = 1/sqrt(2)*norm(Kernel*Distribution{i} - TimeTrace);
      InfluenceMatrix{i} = Kernel*PseudoInverse{i};
      
    %--------------------------------------------------------------------------  
    % Pseudo-Huber (L2-L1) Penalty
    %--------------------------------------------------------------------------  
    case 'Huber'
      
      %Compute approximate optimal Huber Parameter for current value of regularization parameter
      if i>1
        try
        %(Skip the first value since it requires previous unconstrained distributions)
        Solveroptions = optimoptions(@fsolve,'Display','off','Algorithm','trust-region-reflective');
        HuberParameter = fsolve(@(HuberParameter) Kernel'*(Kernel*PseudoInverse{i-1}*TimeTrace - TimeTrace) + ...
          RegularizationParameter(i-1)^2/HuberParameter^2*(L')*((L*PseudoInverse{i-1}*TimeTrace)./sqrt((L*PseudoInverse{i-1}*TimeTrace/HuberParameter).^2 + 1)) ...
          ,options.HuberParameter,Solveroptions);
        % Ensure that Huber parameter does not get negative (can lead to crashes later on)
        HuberParameter = abs(HuberParameter);
        catch
          
        end
      end
      %Unconstrained distributions required for construction of correct pseudoinverse
      Distribution{i} = zeros(DipolarDimension,1);
      try
        %Define options for fsolve
        Solveroptions = optimoptions(@fsolve,'Display','off','Algorithm','trust-region-reflective');
        Distribution{i} = fsolve(@(Distribution) Kernel'*(Kernel*Distribution - TimeTrace) + ...
          RegularizationParameter(i)^2/(HuberParameter^2)*(L')*((L*Distribution)./sqrt((L*(Distribution)/HuberParameter).^2 + 1)) ...
          ,Distribution{i},Solveroptions);
        
        % Compute pseudoinverse by analytical expression (only defined for unconstrained distribution)
        HuberTerm = 1/(HuberParameter^2)*((L)'*(L./sqrt((L*Distribution{i}/HuberParameter).^2 + 1)));
        PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*HuberTerm)\Kernel';
        
        % If constrained solution is requested
        if ~options.unconstrainedProblem
          %Now that we have the pseudoinverse, the constrained distribution can be computed quickly
          Distribution{i} = fnnls(Kernel'*Kernel + RegularizationParameter(i)^2*HuberTerm,Kernel'*TimeTrace);
        end  
      catch
        % Worst-case scenario when fsolve crashes, use this bad method to get at least something
        fprintf('Huber Selection: Failed \n')
        PseudoInverseTemp = (Kernel'*Kernel + RegularizationParameter(i)^2*(L')*L)\Kernel';
        DistributionTemp  = PseudoInverseTemp*TimeTrace;
        HuberTerm = ((L/HuberParameter)'*diag(1./sqrt((L*DistributionTemp/HuberParameter).^2 + 1))*(L/HuberParameter));
        PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*HuberTerm)\Kernel';
        Distribution{i} = fnnls(Kernel'*Kernel + RegularizationParameter(i)^2*HuberTerm,Kernel'*TimeTrace);
      end
      
      %Get last variables required for the selection functionals
      Penalty(i) = sum(sqrt((L*Distribution{i}/HuberParameter).^2 + 1 ) - 1);
      Residual(i) = 1/sqrt(2)*norm(Kernel*Distribution{i} - TimeTrace);
      InfluenceMatrix{i} = Kernel*(PseudoInverse{i});

    %--------------------------------------------------------------------------       
    % Tikhonov (L2) Penalty
    %--------------------------------------------------------------------------       
    otherwise
      PseudoInverse{i} = (Kernel'*Kernel + RegularizationParameter(i)^2*(L')*L)\Kernel';
      if options.unconstrainedProblem
        Distribution{i}  = PseudoInverse{i}*TimeTrace; %unconstrained Tikhonov solution
      else
        Distribution{i} = fnnls(Kernel'*Kernel + RegularizationParameter(i)^2*(L')*L,Kernel'*TimeTrace ); %non-negative Tikhonov solution
      end
      Penalty(i) = 1/sqrt(2)*norm(L*Distribution{i});
      Residual(i) = 1/sqrt(2)*norm(Kernel*Distribution{i} - TimeTrace);
      InfluenceMatrix{i} = Kernel*PseudoInverse{i};
  end
end

%In case variables are in matrix form reshape them to vectors
Distribution = reshape(Distribution,1,[]);
PseudoInverse = reshape(PseudoInverse,1,[]);
Residual = reshape(Residual,1,[]);
InfluenceMatrix = reshape(InfluenceMatrix,1,[]);
Penalty = reshape(Penalty,1,[]);
nPoints = length(Distribution);

%--------------------------------------------------------------------------
% Selection methods for optimal regularization parameter
%--------------------------------------------------------------------------
switch SelectionMethod
  
  case 'lr' %L-curve Minimum-Radius method (LR)
    Eta = log(Penalty);
    Rho = log(Residual);
    Functional = ((Rho - min(Rho))/(max(Rho) - min(Rho))).^2 + ((Eta - min(Eta))/(max(Eta) - min(Eta))).^2;
    Index = find(Functional == min(Functional));
    
  case 'lc' %L-curve Maximum-Curvature method (LC)
    d1Residual = gradient(log(Residual));
    d2Residual = gradient(d1Residual);
    d1Penalty = gradient(log(Penalty));
    d2Penalty = gradient(d1Penalty);
    Functional = (d1Residual.*d2Penalty - d2Residual.*d1Penalty)./(d1Residual.^2 + d1Penalty.^2).^(3/2);
    Index = find(Functional == min(Functional));

  case 'dp' %Discrepancy principle (DP)
    SafetyFactor = 1;
    Noise  = options.noise;
    Index = Residual/sqrt(DipolarDimension) <= SafetyFactor*Noise; 
    Index = RegularizationParameter == max(RegularizationParameter(Index));

  case 'cv' %Cross validation (CV)
    for i=1:nPoints
      InfluenceDiagonal = diag(InfluenceMatrix{i});
      Functional(i) = sum(abs(TimeTrace - Kernel*(Distribution{i})./(ones(DipolarDimension,1) - InfluenceDiagonal)).^2);
    end
    Index = find(Functional == min(Functional));

  case 'gcv' %Generalized Cross Validation (GCV)
    for i=1:nPoints
      Functional(i) = Residual(i)^2/((1 - trace(InfluenceMatrix{i})/DipolarDimension)^2);
    end
    Index = find(Functional == min(Functional));

  case 'rgcv' %Robust Generalized Cross Validation (rGCV)
    TuningParameter = 0.9;
    for i=1:nPoints
      Functional(i) = Residual(i)^2/((1 - trace(InfluenceMatrix{i})/DipolarDimension)^2)*(TuningParameter + (1 - TuningParameter)*trace(InfluenceMatrix{i}^2)/DipolarDimension);
    end
    Index = find(Functional == min(Functional));

  case 'srgcv' %Strong Robust Generalized Cross Validation (srGCV)
    TuningParameter = 0.8;
    for i=1:nPoints
      Functional(i) = Residual(i)^2/((1 - trace(InfluenceMatrix{i})/DipolarDimension)^2)*(TuningParameter + (1 - TuningParameter)*trace(PseudoInverse{i}'*PseudoInverse{i})/DipolarDimension);
    end
    Index = find(Functional == min(Functional));
    
  case 'aic' %Akaike information criterion (AIC)
    Criterion = 2;
    for i=1:nPoints
      Functional(i) = DipolarDimension*log(Residual(i)^2/DipolarDimension) + Criterion*trace(InfluenceMatrix{i});
    end
    Index = find(Functional == min(Functional));
    
  case 'bic' %Bayesian information criterion (BIC)
    Criterion = log(DipolarDimension);
    for i=1:nPoints
      Functional(i) = DipolarDimension*log(Residual(i)^2/DipolarDimension) + Criterion*trace(InfluenceMatrix{i});
    end
    Index = find(Functional == min(Functional));
    
  case 'aicc' %Corrected Akaike information criterion (AICC)
    for i=1:nPoints
      Criterion = 2*DipolarDimension/(DipolarDimension-trace(InfluenceMatrix{i})-1);
      Functional(i) = DipolarDimension*log(Residual(i)^2/DipolarDimension) + Criterion*trace(InfluenceMatrix{i});
    end
    Index = find(Functional == min(Functional));
    
  case 'rm' %Residual method (RM)
    for i=1:nPoints
      Scaling = Kernel'*(eye(size(InfluenceMatrix{i})) - InfluenceMatrix{i});
      Functional(i) = Residual(i)^2/sqrt(trace(Scaling'*Scaling));
    end
    Index = find(Functional == min(Functional));
    
  case 'ee' %Extrapolated Error (EE)
    for i=1:nPoints
      Functional(i) = Residual(i)^2/norm(Kernel'*(Kernel*Distribution{i} - TimeTrace));
    end
    Index = find(Functional == min(Functional));
    
  case 'ncp' %Normalized Cumulative Periodogram (NCP)
    for i=1:nPoints
      ResidualPeriodogram = abs(fft(Kernel*Distribution{i} - TimeTrace)).^2;
      WhiteNoisePowerSpectrum = zeros(length(ResidualPeriodogram),1);
      ResidualPowerSpectrum = zeros(length(ResidualPeriodogram),1);
      for j=1:length(ResidualPeriodogram) - 1
        ResidualPowerSpectrum(j)  = norm(ResidualPeriodogram(2:j+1),1)/norm(ResidualPeriodogram(2:end),1);
        WhiteNoisePowerSpectrum(j) = j/(length(ResidualPeriodogram) - 1);
      end
      Functional(i) = norm(ResidualPowerSpectrum - WhiteNoisePowerSpectrum);
    end
    Index = find(Functional == min(Functional));
    
  case 'gml' %Generalized Maximum Likelihood (GML)
    Treshold = 1e-9;
    for i=1:nPoints
      try %Once crushed beacause of eig(NaN) 
        EigenValues = eig(eye(size(InfluenceMatrix{i})) - InfluenceMatrix{i});
      catch
        EigenValues = 0;
      end
      EigenValues(EigenValues < Treshold) = 0;
      NonZeroEigenvalues = real(EigenValues(EigenValues~=0));
      Functional(i) = TimeTrace'*(TimeTrace - Kernel*Distribution{i})/nthroot(prod(NonZeroEigenvalues),length(NonZeroEigenvalues));
    end
    Index = find(Functional == min(Functional));
    
  case 'mcl' %Mallows' C_L (MCL)
    if ~isfield(options,'NoiseLevel')
      options.NoiseLevel = 0.05;
    end
    NoiseLevel = options.NoiseLevel;
    for i=1:nPoints
      Functional(i) = Residual(i)^2 + 2*NoiseLevel^2*trace(InfluenceMatrix{i}) - 2*DipolarDimension*NoiseLevel^2;
    end
    Index = find(Functional == min(Functional));
    
end

%--------------------------------------------------------------------------
% Finish & Return 
%--------------------------------------------------------------------------

%Show error in case input doesn't correspond to any case
if ~exist('Index','var')
  error('Error: No matching selection method was found. Check your input for options.SelectionMethod')
end

Functional = reshape(Functional,length(HuberParameter),length(RegularizationParameter));
[Row,Column] = find(Functional == min(min(Functional)));
Distribution = reshape(Distribution,length(HuberParameter),length(RegularizationParameter));
try
  OptimalDistribution = Distribution{Row,Column};
  % Extract the found optimal regularization and Huber parameters
  OptimalRegularizationParameter = RegularizationParameter(Column);
  OptimalHuberParameter = HuberParameter(Row);
catch
  fprintf('Something went horribly wrong and selection failed... \n')
  OptimalRegularizationParameter = 1;
  OptimalHuberParameter = 1.35;
end


%In case there should be two points with same minima
if length(OptimalRegularizationParameter) > 1
    OptimalRegularizationParameter = OptimalRegularizationParameter(1);
end
if length(OptimalHuberParameter) > 1
    OptimalHuberParameter = OptimalHuberParameter(1);
end
%In some extreme case that method should crash and no optimal value is found
if isempty(OptimalRegularizationParameter)
  OptimalRegularizationParameter = 1;
end

% Plot the selection method functional and the chosen optimal point
%--------------------------------------------------------------------------
if options.plotSelectionFunctional
  figure(3456),set(gcf,'NumberTitle','off','Name','Regularization Parameter Selection Functional')
  clf,plot(RegularizationParameter, Functional),hold on,plot(OptimalRegularizationParameter,Functional(Index),'ro'),hold off;
  try
  xlim([RegularizationParameter(Index+10), RegularizationParameter(Index - 10)])
  catch 
  end
  if isempty(options.HuberParameter)
  title(sprintf('Optimal lambda = %.3f',OptimalRegularizationParameter))
  else
    title(sprintf('Optimal lambda = %.3f  Optimal Huber = %.3d',OptimalRegularizationParameter,OptimalHuberParameter))
  end
  xlabel('Regularization Parameter \lambda'),ylabel(sprintf('%s - Functional',upper(SelectionMethod)))
  legend({'Functional','Optimum'},'Location','best'),drawnow;
end

% Plot the optimal distance distribution and point-wise Huber penalty
%--------------------------------------------------------------------------
if options.plotL1L2regions && strcmp(options.SelectionPenaltyFunctional,'Huber')
  figure(34567),clf
  Auxiliary = abs(L*OptimalDistribution);
  Axis = 1:numel(OptimalDistribution);
  plot(Axis,OptimalDistribution,'k'),hold on
  p1 = plot(Axis(Auxiliary>OptimalHuberParameter),OptimalDistribution(Auxiliary>OptimalHuberParameter),'ro');
  p2 = plot(Axis(Auxiliary<OptimalHuberParameter),OptimalDistribution(Auxiliary<OptimalHuberParameter),'bo');drawnow
  legend([p1 p2],{'L1','L2'})
  title('L1-L2 Region of optimal distribution')
end

end

